@@ -271,7 +271,6 @@ import tempfile
 import operator
 import shlex
 import traceback
-import types
 
 from ansible.module_utils.compat.version import LooseVersion
 
@@ -459,7 +458,7 @@ def _have_pip_module():  # type: () -> bool
     except ImportError:
         find_spec = None  # type: ignore[assignment] # type: ignore[no-redef]
 
-    if find_spec:
+    if find_spec:  # type: ignore[truthy-function]
         # noinspection PyBroadException
         try:
             # noinspection PyUnresolvedReferences
