@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # -*- coding: utf-8 -*-
 
 # Copyright: (c) 2016, Brian Coca <bcoca@ansible.com>
@@ -9,29 +8,30 @@ __metaclass__ = type
 
 
 DOCUMENTATION = '''
-module: systemd
+module: systemd_service
 author:
     - Ansible Core Team
 version_added: "2.2"
-short_description:  Manage services
+short_description:  Manage systemd units
 description:
-    - Controls systemd services on remote hosts.
+    - Controls systemd units (services, timers, and so on) on remote hosts.
 options:
     name:
         description:
-            - Name of the service. This parameter takes the name of exactly one service to work with.
-            - When using in a chroot environment you always need to specify the full name i.e. (crond.service).
+            - Name of the unit. This parameter takes the name of exactly one unit to work with.
+            - When no extension is given, it is implied to a C(.service) as systemd.
+            - When using in a chroot environment you always need to specify the name of the unit with the extension. For example, C(crond.service).
         type: str
         aliases: [ service, unit ]
     state:
         description:
             - C(started)/C(stopped) are idempotent actions that will not run commands unless necessary.
-              C(restarted) will always bounce the service. C(reloaded) will always reload.
+              C(restarted) will always bounce the unit. C(reloaded) will always reload.
         type: str
         choices: [ reloaded, restarted, started, stopped ]
     enabled:
         description:
-            - Whether the service should start on boot. B(At least one of state and enabled are required.)
+            - Whether the unit should start on boot. B(At least one of state and enabled are required.)
         type: bool
     force:
         description:
@@ -45,7 +45,7 @@ options:
     daemon_reload:
         description:
             - Run daemon-reload before doing any other operations, to make sure systemd has read any changes.
-            - When set to C(yes), runs daemon-reload even if the module does not start or stop anything.
+            - When set to C(true), runs daemon-reload even if the module does not start or stop anything.
         type: bool
         default: no
         aliases: [ daemon-reload ]
@@ -63,7 +63,7 @@ options:
             - "For systemd to work with 'user', the executing user must have its own instance of dbus started and accessible (systemd requirement)."
             - "The user dbus process is normally started during normal login, but not during the run of Ansible tasks.
               Otherwise you will probably get a 'Failed to connect to bus: no such file or directory' error."
-            - The user must have access, normally given via setting the ``XDG_RUNTIME_DIR`` variable, see example below.
+            - The user must have access, normally given via setting the C(XDG_RUNTIME_DIR) variable, see example below.
 
         type: str
         choices: [ system, user, global ]
@@ -76,18 +76,28 @@ options:
         type: bool
         default: no
         version_added: "2.3"
+extends_documentation_fragment: action_common_attributes
+attributes:
+    check_mode:
+        support: full
+    diff_mode:
+        support: none
+    platform:
+        platforms: posix
 notes:
-    - Since 2.4, one of the following options is required 'state', 'enabled', 'masked', 'daemon_reload', ('daemon_reexec' since 2.8),
-      and all except 'daemon_reload' (and 'daemon_reexec' since 2.8) also require 'name'.
-    - Before 2.4 you always required 'name'.
-    - Globs are not supported in name, i.e ``postgres*.service``.
-    - Supports C(check_mode).
+    - Since 2.4, one of the following options is required C(state), C(enabled), C(masked), C(daemon_reload), (C(daemon_reexec) since 2.8),
+      and all except C(daemon_reload) and (C(daemon_reexec) since 2.8) also require C(name).
+    - Before 2.4 you always required C(name).
+    - Globs are not supported in name, i.e C(postgres*.service).
+    - The service names might vary by specific OS/distribution
+    - The order of execution when having multiple properties is to first enable/disable, then mask/unmask and then deal with service state.
+      It has been reported that systemctl can behave differently depending on the order of operations if you do the same manually.
 requirements:
     - A system managed by systemd.
 '''
 
 EXAMPLES = '''
-- name: Make sure a service is running
+- name: Make sure a service unit is running
   ansible.builtin.systemd:
     state: started
     name: httpd
@@ -100,33 +110,33 @@ EXAMPLES = '''
 - name: Restart service cron on centos, in all cases, also issue daemon-reload to pick up config changes
   ansible.builtin.systemd:
     state: restarted
-    daemon_reload: yes
+    daemon_reload: true
     name: crond
 
 - name: Reload service httpd, in all cases
   ansible.builtin.systemd:
-    name: httpd
+    name: httpd.service
     state: reloaded
 
 - name: Enable service httpd and ensure it is not masked
   ansible.builtin.systemd:
     name: httpd
-    enabled: yes
+    enabled: true
     masked: no
 
-- name: Enable a timer for dnf-automatic
+- name: Enable a timer unit for dnf-automatic
   ansible.builtin.systemd:
     name: dnf-automatic.timer
     state: started
-    enabled: yes
+    enabled: true
 
 - name: Just force systemd to reread configs (2.4 and above)
   ansible.builtin.systemd:
-    daemon_reload: yes
+    daemon_reload: true
 
 - name: Just force systemd to re-execute itself (2.8 and above)
   ansible.builtin.systemd:
-    daemon_reexec: yes
+    daemon_reexec: true
 
 - name: Run a user service when XDG_RUNTIME_DIR is not set on remote login
   ansible.builtin.systemd:
@@ -139,7 +149,7 @@ EXAMPLES = '''
 
 RETURN = '''
 status:
-    description: A dictionary with the key=value pairs returned from `systemctl show`.
+    description: A dictionary with the key=value pairs returned from C(systemctl show).
     returned: success
     type: complex
     sample: {
@@ -359,7 +369,7 @@ def main():
     if os.getenv('XDG_RUNTIME_DIR') is None:
         os.environ['XDG_RUNTIME_DIR'] = '/run/user/%s' % os.geteuid()
 
-    ''' Set CLI options depending on params '''
+    # Set CLI options depending on params
     # if scope is 'system' or None, we can ignore as there is no extra switch.
     # The other choices match the corresponding switch
     if module.params['scope'] != 'system':
@@ -383,13 +393,19 @@ def main():
     if module.params['daemon_reload'] and not module.check_mode:
         (rc, out, err) = module.run_command("%s daemon-reload" % (systemctl))
         if rc != 0:
-            module.fail_json(msg='failure %d during daemon-reload: %s' % (rc, err))
+            if is_chroot(module) or os.environ.get('SYSTEMD_OFFLINE') == '1':
+                module.warn('daemon-reload failed, but target is a chroot or systemd is offline. Continuing. Error was: %d / %s' % (rc, err))
+            else:
+                module.fail_json(msg='failure %d during daemon-reload: %s' % (rc, err))
 
     # Run daemon-reexec
     if module.params['daemon_reexec'] and not module.check_mode:
         (rc, out, err) = module.run_command("%s daemon-reexec" % (systemctl))
         if rc != 0:
-            module.fail_json(msg='failure %d during daemon-reexec: %s' % (rc, err))
+            if is_chroot(module) or os.environ.get('SYSTEMD_OFFLINE') == '1':
+                module.warn('daemon-reexec failed, but target is a chroot or systemd is offline. Continuing. Error was: %d / %s' % (rc, err))
+            else:
+                module.fail_json(msg='failure %d during daemon-reexec: %s' % (rc, err))
 
     if unit:
         found = False
@@ -488,11 +504,15 @@ def main():
 
             # do we need to enable the service?
             enabled = False
-            (rc, out, err) = module.run_command("%s is-enabled '%s'" % (systemctl, unit))
+            (rc, out, err) = module.run_command("%s is-enabled '%s' -l" % (systemctl, unit))
 
             # check systemctl result or if it is a init script
             if rc == 0:
                 enabled = True
+                # Check if the service is indirect or alias and if out contains exactly 1 line of string 'indirect'/ 'alias' it's disabled
+                if out.splitlines() == ["indirect"] or out.splitlines() == ["alias"]:
+                    enabled = False
+
             elif rc == 1:
                 # if not a user or global user service and both init script and unit file exist stdout should have enabled/disabled, otherwise use rc entries
                 if module.params['scope'] == 'system' and \
